#!/usr/bin/env vpython3
# Copyright 2024 The Chromium Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.
"""Script to extract edits from clang spanification tool output.

The edits have the following format:
    ...
    e{lhs_node1}@{rhs_node2}           # Edge from node 1 to node 2.
    e{lhs_node2}@{rhs_node3}           # Edge from node 2 to node 3.
    e{lhs_node3}@{rhs_node4}           # Edge from node 3 to node 4.
    ...
    s{node_1}                          # Source node of the graph that triggers
                                       # a rewrite (i.e. buffer usage)
    ...
    f{lhs_key}@{rhs_key}@{replacement} # Span frontier replacement applied if
                                       # lhs_key is not rewritten but rhs_key
                                       # is.
    ...
Where lhs_node, rhs_node, and node_n represent a node's text representation
generated using the spanification tool's Node::ToString() function.

A node has the following format:
size_info_available\,is_dependent,(\,replacement)+

size_info_available: 0 or 1
is_dependent: 0 or 1

extract_edits.py takes input that is concatenated from multiple tool
invocations and extract just the edits with the following steps:
1- Construct the adjacency list of nodes
   (a pairs of nodes represents an edge in the directed graph)

2- Determine whether size info is available for a given source node.

3- Run `DFS` starting from source nodes whose size info is available and emit
   edits for reachable nodes.

4- Adapt dereference expressions and add data changes where necessary.

extract_edits.py would then emit the following output:
    <edit1>
    <edit2>
    <edit3>
    ...
Where the edit is either a replacement or an include directive.

For more details about how the tool works, see the doc here:
https://docs.google.com/document/d/1hUPe21CDdbT6_YFHl03KWlcZqhNIPBAfC-5N5DDY2OE/
"""

import sys
import urllib.parse

from os.path import expanduser
import pprint


# The connected components in the graph. This is useful to split the rewrite
# into atomic changes.
class Component:
    all = set()

    def __init__(self) -> None:
        # Changes associated with the connected component.
        self.changes = set()

        # Frontier changes are either accepted or rejected. The two dictionaries
        # are used to detect conflicts in the frontier changes. This might
        # happen in rare cases where C++ macros are used. In case of conflict,
        # the whole component is discarded.
        self.frontier_changes_accepted = set()
        self.frontier_changes_rejected = set()

        # `Component.all` can be used to iterate over all components.
        Component.all.add(self)

class Node:
    # Mapping in between the node's key and the node.
    key_to_node = dict()

    def __init__(self, size_info_available, is_dependent,
                 *replacements) -> None:
        self.replacements = replacements
        for replacement in replacements:
            assert_valid_replacement(replacement)

        self.is_dependent = is_dependent

        # Neighbors of the node in the graph. The graph is directed,
        # flowing from lhs to rhs.
        self.neighbors_directed = set()
        self.neighbors_undirected = set()

        # Property to track whether the node is "connected" to a source node.
        # This is set from DFS(...)
        self.visited = False

        # See SizeInfoAvailable(...) for more details.
        self.size_info_available = size_info_available
        self.size_info_step = False

        # Identify the connected component this node belongs to. This is set in
        # the main function.
        self.component = None

    # The key of the node is the first replacement.
    def key(self) -> str:
        return self.replacements[0]

    def __eq__(self, other):
        if isinstance(other, Node):
            return self.key() == other.key()
        return False

    def __hash__(self) -> int:
        return hash(self.key())

    # Static method to get a node from a replacement key.
    @classmethod
    def from_key(cls: type, replacement: str):
        return cls.key_to_node.get(replacement)

    # Static method to create a node from its string representation. This
    # deduplicate nodes by storing them in a dictionary.
    @classmethod
    def get_or_create(cls: type, txt: str):
        x = txt.split('\\,')

        # Expect exactly 6 elements that correspond to the following node
        # attributes:
        # - size_info_available
        # - is_dependent
        # - replacements+
        assert len(x) >= 3, txt

        # Value are escaped to avoid conflicts with the separator. Unescape
        # them.
        x = [urllib.parse.unquote(y) for y in x]

        # `./apply-edits.py` expects `\n` to be escaped.
        x = [y.replace('\n', '\0') for y in x]

        node = Node(*x)

        # Deduplicate nodes, as they might appear multiple times in the input.
        if (Node.key_to_node.get(node.key()) is None):
            Node.key_to_node[node.key()] = node

        return Node.key_to_node[node.key()]

    def __repr__(self) -> str:
        result = [
            f"Node {hash(self)} {{",
            f"  key: {self.key()}",
            f"  size_info_available: {self.size_info_available}",
            f"  neighbors_directed: {pprint.pformat([hash(n) for n in self.neighbors_directed], indent=4)}",
            "}",
        ]
        return "\n".join(result)

    # This is not parsable by get_or_create but is useful for debugging the
    # graph of nodes.
    def to_debug_string(self) -> str:
        return repr(self)

    # Static method to get all nodes.
    @classmethod
    def all(cls: type):
        return cls.key_to_node.values()


def DFS(node: Node):
    """
    Explore the graph in depth-first search from the given node. Identify edits
    to apply.

    Args:
        node: The current node being processed.
    """
    # Only visit nodes once:
    if (node.visited):
        return
    node.visited = True

    if not node.key().endswith('<empty>'):
        for replacement in node.replacements:
            node.component.changes.add(replacement)

    for neighbour in node.neighbors_directed:
        DFS(neighbour)


def SizeInfoAvailable(node: Node):
    """
    Determines whether size information is available for a source node and its
    neighbors_directed. Updates the node's size_info_available attribute.

    Args:
        node: The current node's being processed.

    Returns:
        None if a cycle is detected,
        '1' if size information is available for the node and its neighbors,
        '0' otherwise,
    """

    # If we reached a node that contains size info, just return that.
    if node.size_info_available == '1':
        return '1'

    # Cycle detection: If the node is currently being visited, there's a cycle.
    # We can't determine the size info for this node.
    if node.size_info_step == 'visiting':
        return None

    # Memoization: If the node has already been visited, return the result
    # immediately.
    if node.size_info_step == 'visited':
        return node.size_info_available

    node.size_info_step = 'visiting'
    size_info_available = '0'
    # Check neighbors. If any neighbor doesn't have size info or there's a
    # cycle, the current node also doesn't.
    for neighbour in node.neighbors_directed:
        # Break as soon as we encounter a neighbor for which size info is not
        # available.
        if SizeInfoAvailable(neighbour) == '0':
            size_info_available = '0'
            break
        size_info_available = '1'

    node.size_info_available = size_info_available
    node.size_info_step = 'visited'
    return size_info_available


# Assert a replacement follows the expected format:
# - r:::<file path>:::<offset>:::<length>:::<replacement text>
# - include-user-header:::<file path>:::-1:::-1:::<include text>
# - include-system-header:::<file path>:::-1:::-1:::<include text>
def assert_valid_replacement(replacement: str):
    try:
        parts = replacement.split(':::')
        assert len(parts) == 5
        assert parts[0] in [
            'r', 'include-user-header', 'include-system-header'
        ]
        assert parts[1] != ''  # File path
        int(parts[2].isdigit())  # Offset
        int(parts[3].isdigit())  # Length
    except:
        # Augment the error with the replacement text for better debugging.
        assert False, f"Invalid replacement: \"{replacement}\""

def main():
    # Since the tool is invoked from multiple compile units, we are using sets
    # to deduplicate what was visible from multiple compile units.

    # A set of source nodes that trigger the rewrite.
    sources = set()

    # Change to apply at the edge in between rewritten and non-rewritten nodes.
    frontiers = set()

    # Collect from every compile units the nodes and edges of the graph:
    for line in sys.stdin:
        line = line.rstrip('\n\r')

        # The first character of the line denotes the type of the line:
        # - 'e': Edge in between two nodes.
        # - 's': Source node of the graph triggering the rewrite.
        # - 'f': Span frontier change.
        assert line[0] in ['e', 's', 'f'], "Unknown line type: " +\
               line[0] + " in line: " + line

        # Source node:
        if line[0] == 's':
            assert_valid_replacement(line[1:])
            sources.add(line[1:])
            continue

        # Edge in between two nodes:
        if line[0] == 'e':
            nodes = line[1:].split('@')
            assert len(nodes) == 2, "Invalid edge: " + line
            lhs = Node.get_or_create(nodes[0])
            rhs = Node.get_or_create(nodes[1])

            # Directed edge:
            lhs.neighbors_directed.add(rhs)

            # Undirected edge:
            lhs.neighbors_undirected.add(rhs)
            rhs.neighbors_undirected.add(lhs)
            continue

        # Span frontier change:
        if line[0] == 'f':
            frontiers.add(line[1:])
            continue

        assert False, "Unreachable code"

    # Mark the source nodes:
    source_nodes = []
    for source in sources:
        source_node = Node.from_key(source)
        # When using templates, it is possible the source isn't part of any
        # edges. In this case, it can't be rewritten, because we don't know
        # if its size info is available.
        #
        # This is a limitation of the current implementation. We could improve
        # this by adding a new line that indicates whether the node's size
        # information is available. It shouldn't be part of the edges.
        if source_node is None:
            continue
        source_nodes.append(source_node)

        # Determine whether size information is available from this source.
        SizeInfoAvailable(source_node)

    # Identify all the connected components in the undirected graph. This is
    # exploring the graph in depth-first search and assigning the same component
    # to each node in the connected component.
    for node in Node.all():
        if node.component is not None:
            continue
        new_component = Component()
        stack = [node]
        while stack:
            current = stack.pop()
            if current.component is not None:
                continue
            current.component = new_component
            for neighbor in current.neighbors_undirected:
                stack.append(neighbor)

    # Collect the changes to apply. Starting from sources nodes whose size info
    # could be determined.
    for node in source_nodes:
        # Some sources might not have their size info available. We can't
        # rewrite those.
        if node.size_info_available != '1':
            continue

        # Collect the changes to apply. We start from sources nodes whose size
        # info is available and explore the graph in depth-first search.
        DFS(node)

    # Iterate over the dependent nodes and then check if their only neighbor was
    # visited. Visited nodes here are nodes who's type was rewritten to span.
    # In that case, the dependent expression needs to be adapted/rewritten.
    for node in Node.all():
        if node.is_dependent == '1':
            neighbor = list(node.neighbors_directed)[0]
            if neighbor.visited:
                for replacement in node.replacements:
                    neighbor.component.changes.add(replacement)

    # At the edge in between rewritten and non-rewritten nodes, we need
    # to add a call to `.data()` to access the pointer from the span:
    for frontier in frontiers:
        (lhs_key, rhs_key, replacement) = frontier.split('@')
        lhs_node = Node.from_key(lhs_key)
        rhs_node = Node.from_key(rhs_key)

        apply_frontier = rhs_node.visited and not lhs_node.visited
        if apply_frontier:
            lhs_node.component.frontier_changes_accepted.add(replacement)
        else:
            lhs_node.component.frontier_changes_rejected.add(replacement)

    # Do or do not, there is no try. Discard components with conflicting
    # frontier changes. This happens in rare cases where C++ macros are used.
    # The whole component is discarded in case of conflict, because we can't
    # satisfy the constraints.
    for component in Component.all:
        if component.frontier_changes_accepted & component.frontier_changes_rejected:
            component.changes.clear()
            continue;

        component.changes |= component.frontier_changes_accepted

    # Emit the changes:
    # - ~/scratch/patches.txt: A summary of each atomic change.
    # - ~/scratch/patch_<patch_index>: Write each atomic change.
    # - stdout: Print a bundle of all the changes. This is usually piped to
    #           "./tools/clang/scripts/apply_edits.py" to apply the changes.

    summary_filename = expanduser('~/scratch/patches.txt')
    summary_file = open(summary_filename, 'w')

    component_with_changes = [
        component for component in Component.all if len(component.changes) > 0
    ]

    for index, component in enumerate(component_with_changes):
        for text in component.changes:
            print(text)

        summary_file.write(f'patch_{index}: {len(component.changes)}\n')

        with open(expanduser(f'~/scratch/patch_{index}.txt'), 'w') as f:
            f.write('\n'.join(component.changes))

    summary_file.close()

    return 0


if __name__ == '__main__':
    sys.exit(main())
